<?php

namespace Drupal\nft_metadata;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Link;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Psr\Log\LoggerInterface;

/**
 * Manipulates entity type to rule the world.
 */
class EntityTypeInfo {

  use StringTranslationTrait;

  /**
   * A logger channel.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity type bundle info service.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected $bundleInfoService;

  /**
   * Construct a new object.
   */
  public function __construct(
    LoggerInterface $logger,
    EntityTypeManagerInterface $entity_type_manager,
    EntityTypeBundleInfoInterface $entity_type_bundle_info
  ) {
    $this->logger = $logger;
    $this->entityTypeManager = $entity_type_manager;
    $this->bundleInfoService = $entity_type_bundle_info;

  }

  /*************************************************************************
   * Hooks.
   *************************************************************************/

  /**
   * Alter types to set custom link template to entities' metadata.
   *
   * @param array &$entity_types
   *   List of entity types, passed by reference
   */
  public function entityTypeAlter(array &$entity_types) {
    foreach ($entity_types as $entity_type_id => $entity_type) {
      if ($entity_type->entityClassImplements(ContentEntityInterface::class)) {
        // Register entity metadata route link.
        $base_path = $entity_type->hasLinkTemplate('canonical') ? $entity_type->getLinkTemplate('canonical') : NULL;
        $base_path = $base_path ?: ($entity_type->hasLinkTemplate('edit-form') ? $entity_type->getLinkTemplate('edit-form') : NULL);
        $base_path = $base_path ?: "/$entity_type_id/{$entity_type_id}";
        $entity_type->setLinkTemplate('nft-metadata', $base_path . '/nft/metadata');
      }
    }
  }

  /**
   * Adds nft_metadata operation on entity that supports it.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity on which to define an operation.
   *
   * @return array
   *   An array of operation definitions.
   *
   * @see nft_metadata_entity_operation()
   */
  public function entityOperation(EntityInterface $entity) {
    $operations = [];

    // Display link to NFT Metadata operation, if possible.
    // @see nft_metadata_entity_access();
    // @see \Drupal\nft_metadata\EntityTypeInfo::entityOperationAccess();
    if ($entity->hasLinkTemplate('nft-metadata') && $entity->access('nft-metadata')) {
      $operations['nft_metadata'] = [
        'title' => $this->t('Metadata'),
        'weight' => 100,
        'url' => $entity->toUrl('nft-metadata'),
      ];
    }

    return $operations;
  }

  /**
   * Implements a hook bridge for hook_entity_access().
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to check access for.
   * @param string $operation
   *   The operation being performed.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The user account making the to check access for.
   *
   * @return \Drupal\Core\Access\AccessResult
   *   The result of the access check.
   *
   * @see nft_metadata_entity_access()
   *
   * @todo Implements dynamic permissions?
   */
  public function entityOperationAccess(EntityInterface $entity, $operation, AccountInterface $account) {
    if ($operation == 'nft-metadata') {
      $entity_type = $entity->getEntityType();
      $entity_type_id = $entity_type->id();
      $bundle = $entity_type->hasKey('bundle') ? $entity->bundle() : NULL;
      $permission = $bundle ? "access $entity_type_id $bundle metadata" : "access $entity_type_id metadata";

      return AccessResult::allowedIf($account->hasPermission($permission));
    }

    return AccessResult::neutral();
  }

  /*************************************************************************
   * Callbacks.
   *************************************************************************/

  /**
   * Builds a list of custom permissions for this module.
   *
   * @return array
   *   The permissions.
   */
  public function buildPermissions() {
    $permissions = [];
    foreach ($this->entityTypeManager->getDefinitions() as $entity_type) {
      if (!$entity_type->entityClassImplements(ContentEntityInterface::class)) {
        continue;
      }
      if (!$entity_type->hasLinkTemplate('nft-metadata')) {
        continue;
      }

      $entity_type_id = $entity_type->id();
      if (!$entity_type->hasKey('bundle')) {
        // Access to "nft-metadata" operation.
        $permissions["access $entity_type_id metadata"] = [
          'title' => $this->t('Access @label metadata', ['@label' => $entity_type->getSingularLabel()]),
          'description' => $this->t('Provides the ability to access "NFT Metadata" form.'),
          'restrict access' => TRUE,
          'provider' => 'nft_metadata',
        ];

        continue;
      }

      $bundles = $this->bundleInfoService->getBundleInfo($entity_type_id);
      foreach ($bundles as $bundle_id => $bundle_info) {
        // Access to "nft-metadata" operation, by bundle.
        $permissions["access $entity_type_id $bundle_id metadata"] = [
          'title' => $this->t('Access @bundle (@label) metadata', [
            '@label' => ucfirst($entity_type->getSingularLabel()),
            '@bundle' => $bundle_info['label'],
          ]),
          'description' => $this->t('Provides the ability to access "NFT Metadata" form.'),
          'restrict access' => TRUE,
          'provider' => 'nft_metadata',
        ];
      }
    }
    return $permissions;
  }

}
