<?php

namespace Drupal\nft_metadata\Normalizer;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\serialization\Normalizer\NormalizerBase;
use Psr\Log\LoggerInterface;

/**
 * Converts NFT from/to Entity metadata.
 */
class MetadataNormalizer extends NormalizerBase {

  /**
   * The logger channel.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Constructs a MetadataNormalizer object.
   *
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger channel.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(
    LoggerInterface $logger,
    ModuleHandlerInterface $module_handler,
    ConfigFactoryInterface $config_factory,
    EntityTypeManagerInterface $entity_type_manager
  ) {
    $this->logger = $logger;
    $this->moduleHandler = $module_handler;
    $this->configFactory = $config_factory;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public function supportsNormalization($data, $format = NULL, array $context = []) {
    if (!is_object($data) || !$this->checkFormat($format)) {
      return FALSE;
    }

    // Data can only be generated out of a content.
    return $data instanceof ContentEntityInterface;
  }

  /**
   * {@inheritdoc}
   */
  public function normalize($entity, $format = NULL, array $context = []) {
    $entity_uuid = $entity->uuid();
    $entity_type = $entity->getEntityType();
    $entity_type_id = $entity->getEntityTypeId();
    $entity_bundle = $entity_type->hasKey('bundle') ? $entity->bundle() : $entity_type_id;
    $entity_url = $entity->hasLinkTemplate('canonical') ? $entity->toUrl('canonical')
      ->setAbsolute(TRUE)
      ->toString() : NULL;

    $storage = $this->entityTypeManager->getStorage($entity_type_id);

    // Allow other modules to alter contextual values before normalization.
    $this->moduleHandler->alter('nft_metadata_context', $entity, $context);

    // Basic required information.
    // @code
    //  {
    //    "name": "Dave Starbelly",
    //    "description": "Friendly OpenSea Creature that enjoys long swims in the ocean.",
    //    "external_url": "https://openseacreatures.io/3",
    //    "image": "https://storage.googleapis.com/opensea-prod.appspot.com/puffs/3.png",
    //    "attributes": [ ... ],
    //  }
    // @see https://docs.opensea.io/docs/metadata-standards#metadata-structure
    $values = [];
    $values['name'] = $context['name'] ?? $entity->label();
    $values['description'] = $context['description'] ?? t('An @entity_type from @site', [
      '@entity_type' => $entity_type->getSingularLabel(),
      '@site' => $this->configFactory->get('system.site')->get('name'),
    ]);
    $values['external_url'] = $context['external_url'] ?? $entity_url;
    $values['image_url'] = $context['image_url'] ?? NULL;

    // Custom fields.
    $values['custom_fields'] = [
      'entity_type' => $entity_type_id,
      'entity_uuid' => $entity_uuid,
    ];
    if ($revision_id = $storage->getLatestRevisionId($entity->id())) {
      $values['custom_fields']['revision_id'] = $revision_id;
    }

    // Attributes.
    $values['attributes'] = [];
    foreach ($entity_type->getKeys() as $key) {
      if (in_array($key, ['uid'])) {continue;}
      $values['attributes'][$key] = $entity->get($key)->getString();
    }

    // Publicly accessible file holding NFT metadata.
    $values['file_url'] = $context['file_url'] ?? (
      $entity->hasLinkTemplate('nft-metadata') ? $entity->toUrl('nft-metadata')
        ->setAbsolute(TRUE)
        ->setOption('query', ['_format' => 'json'])
        ->toString() : NULL
    );

    // Allow other modules to alter metadata.
    $this->moduleHandler->alter('nft_metadata_values', $values, $entity, $context);

    // Other properties used by some APIs.
    // OpenSea.io and NFTPort.xyz expect `image_url` or `file_url` but others
    // services - such as NFTStorage.io - expects `image` or `file`, without `_url`.
    $values['image'] = $values['image_url'];
    $values['file'] = $values['file_url'];

    return $values;
  }

}
