<?php

namespace Drupal\nft_mint;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\Entity\ConfigEntityBundleBase;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityFormInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Plugin\PluginWithFormsInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\nft_mint\Form\EntityMintForm;
use Drupal\nft_mint\Minter\MinterManagerInterface;
use Psr\Log\LoggerInterface;

/**
 * Manipulates entity type to rule the world.
 */
class EntityTypeInfo {

  use StringTranslationTrait;

  /**
   * A logger channel.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * The config factory service.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity type bundle info service.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected $bundleInfoService;

  /**
   * The Minter plugin manager.
   *
   * @var \Drupal\nft_mint\Minter\MinterManagerInterface
   */
  protected $minterManager;

  /**
   * Construct a new object.
   */
  public function __construct(
    LoggerInterface $logger,
    AccountInterface $current_user,
    ConfigFactoryInterface $config_factory,
    ModuleHandlerInterface $module_handler,
    EntityTypeManagerInterface $entity_type_manager,
    EntityTypeBundleInfoInterface $entity_type_bundle_info,
    MinterManagerInterface $plugin_manager_minter
  ) {
    $this->logger = $logger;
    $this->currentUser = $current_user;
    $this->configFactory = $config_factory;
    $this->moduleHandler = $module_handler;
    $this->entityTypeManager = $entity_type_manager;
    $this->bundleInfoService = $entity_type_bundle_info;
    $this->minterManager = $plugin_manager_minter;
  }

  /*************************************************************************
   * Hooks.
   *************************************************************************/

  /**
   * Alter types to set custom link template on "mintable" entities.
   *
   * @param array &$entity_types
   *   List of entity types, passed by reference
   */
  public function entityTypeAlter(array &$entity_types) {
    foreach ($entity_types as $entity_type_id => $entity_type) {
      if ($entity_type->entityClassImplements(ContentEntityInterface::class)) {
        // Register entity mint route link.
        $base_path = $entity_type->hasLinkTemplate('canonical') ? $entity_type->getLinkTemplate('canonical') : NULL;
        $base_path = $base_path ?: ($entity_type->hasLinkTemplate('edit-form') ? $entity_type->getLinkTemplate('edit-form') : NULL);
        $base_path = $base_path ?: "/$entity_type_id/{$entity_type_id}";
        $entity_type->setLinkTemplate('nft-mint', $base_path . '/nft/mint');

        // Register entity mint base form class.
        if (!$entity_type->getFormClass('nft_mint')) {
          $entity_type->setFormClass('nft_mint', EntityMintForm::class);
        }
      }
    }
  }

  /**
   * Alters bundle forms to set our custom third party settings.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param string $form_id
   *   The form id.
   *
   * @see nft_mint_form_alter()
   */
  public function formAlter(array &$form, FormStateInterface $form_state, $form_id) {
    $form_object = $form_state->getFormObject();

    // Applies to configuration entity types only.
    if (!$form_object instanceof EntityFormInterface || !$form_object->getEntity() instanceof ConfigEntityBundleBase) {
      return;
    }

    $config_entity = $form_state->getFormObject()->getEntity();
    $form_state->set('third_party_settings', $config_entity->getThirdPartySettings('nft_mint') ?? []);

    // Display third party settings.
    $form['nft'] = $form['nft'] ?? [
      '#type' => 'details',
      '#title' => $this->t('NFT settings'),
      '#group' => 'additional_settings',
    ];

    // Minter settings.
    $this->buildMintableElements($form['nft'], $form_state);

    // IPFS settings.
    if ($this->moduleHandler->moduleExists('ipfs')) {
      $this->buildIpfsElements($form['nft'], $form_state);
    }

    // Save third party settings.
    $form['#entity_builders'][] = [static::class, 'entityFormEntityTypeBuild'];
  }

  /**
   * Adds nft_mint operation on entity that supports it.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity on which to define an operation.
   *
   * @return array
   *   An array of operation definitions.
   *
   * @see nft_mint_entity_operation()
   */
  public function entityOperation(EntityInterface $entity) {
    $operations = [];

    // Display link to NFT mint operation, if possible.
    // @see nft_mint_entity_access();
    // @see \Drupal\nft_mint\EntityTypeInfo::entityOperationAccess();
    if ($entity->hasLinkTemplate('nft-mint') && $entity->access('nft-mint')) {
      $operations['nft_mint'] = [
        'title' => $this->t('Mint'),
        'weight' => 100,
        'url' => $entity->toUrl('nft-mint'),
      ];
    }

    return $operations;
  }

  /**
   * Implements a hook bridge for hook_entity_access().
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to check access for.
   * @param string $operation
   *   The operation being performed.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The user account making the to check access for.
   *
   * @return \Drupal\Core\Access\AccessResult
   *   The result of the access check.
   *
   * @see nft_mint_entity_access()
   *
   * @todo Implements dynamic permissions?
   */
  public function entityOperationAccess(EntityInterface $entity, $operation, AccountInterface $account) {
    if ($operation == 'nft-mint') {
      $entity_type = $entity->getEntityType();
      $entity_type_id = $entity_type->id();
      $bundle = $entity_type->hasKey('bundle') ? $entity->bundle() : NULL;
      $mintable_permission = $bundle ? "mint $entity_type_id $bundle entity" : "mint $entity_type_id entity";
      return AccessResult::forbiddenIf(!$this->isMintable($entity, $account) || !$account->hasPermission($mintable_permission));
    }

    return AccessResult::neutral();
  }

  /*************************************************************************
   * Callbacks.
   *************************************************************************/

  /**
   * Builds a list of custom permissions for this module.
   *
   * @return array
   *   The permissions.
   */
  public function buildPermissions() {
    $permissions = [];
    foreach ($this->entityTypeManager->getDefinitions() as $entity_type) {
      if (!$entity_type->entityClassImplements(ContentEntityInterface::class)) {
        continue;
      }
      if (!$entity_type->hasLinkTemplate('nft-mint')) {
        continue;
      }

      $entity_type_id = $entity_type->id();
      if (!$entity_type->hasKey('bundle')) {
        // Access to "nft-mint" operation.
        $permissions["mint $entity_type_id entity"] = [
          'title' => $this->t('Mint @label as NFT', ['@label' => $entity_type->getSingularLabel()]),
          'description' => $this->t('Provides the ability to access "NFT Mint" form.'),
          'restrict access' => TRUE,
          'provider' => 'nft_mint',
        ];

        continue;
      }

      $bundles = $this->bundleInfoService->getBundleInfo($entity_type_id);
      foreach ($bundles as $bundle_id => $bundle_info) {
        // Access to "nft-mint" operation, by bundle.
        $permissions["mint $entity_type_id $bundle_id entity"] = [
          'title' => $this->t('Mint NFT from @bundle (@label)', [
            '@label' => ucfirst($entity_type->getSingularLabel()),
            '@bundle' => $bundle_info['label'],
          ]),
          'description' => $this->t('Provides the ability to access "NFT Mint" form.'),
          'restrict access' => TRUE,
          'provider' => 'nft_mint',
        ];
      }
    }
    return $permissions;
  }

  /**
   * Entity builder method for Entity Type type.
   *
   * @code
   * third_party_settings:
   *   nft_mint:
   *     mint:
   *       mintable: 1
   *       minter: nftport_easy_minting
   *     ipfs:
   *       provider: localhost
   *       auto: 1
   * @endcode
   *
   * @param string $type_id
   *   The type of the entity.
   * @param object $type
   *   The entity type form being built.
   * @param array $form
   *   A nested array form elements comprising the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @see \Drupal\nft_mint\EntityTypeInfo::entityFormAlter()
   */
  public static function entityFormEntityTypeBuild($type_id, $type, array $form, FormStateInterface $form_state) {
    foreach (['nft_mint', 'nft_ipfs'] as $wrapper) {
      $values = [];
      foreach ($form_state->getValues() as $key => $value) {
        if (strpos($key, $wrapper) === 0) {
          $clean_key = str_replace($wrapper . '_', '', $key);
          $values[$clean_key] = $value;
        }
      }

      // Clean wrapper key for easier use later.
      $clean_wrapper = str_replace('nft_', '', $wrapper);

      // Save array of values for each wrapper.
      $type->setThirdPartySetting('nft_mint', $clean_wrapper, $values);
    }
  }

  /*************************************************************************
   * Utility methods.
   *************************************************************************/

  /**
   * Check if an entity type is mintable.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   A given entity.
   *
   * @return bool
   */
  public function isMintable(EntityInterface $entity, AccountInterface $account = NULL) {
    $settings = $this->getThirdPartySettings($entity);
    $is_mintable = $settings['mint']['mintable'] ?? FALSE;
    if (!$account) {return (bool) $is_mintable;}
    $can_mint = $account->hasPermission('mint ' . $entity->getEntityTypeId() . ' entity');
    return (bool) ($is_mintable && $can_mint);
  }

  /**
   * Load nft_mint third party settings from configuration.
   *
   * @param \Drupal\Core\EntityInterface $entity
   *   A given entity ID.
   *
   * @return array
   *   Always return an array. Might be empty.
   */
  public function getThirdPartySettings(EntityInterface $entity) {
    $entity_type = $entity->getEntityType();
    $bundle_id = $entity_type->hasKey('bundle') ? $entity->bundle() : $entity_type->id();
    $config_key = $entity_type->id() . '.type.' . $bundle_id;
    if ($config = $this->configFactory->get($config_key)) {
      return $config->get('third_party_settings')['nft_mint'] ?? [];
    }

    return [];
  }

  /**
   * Get a minter plugin form.
   */
  public static function getPluginForm($plugin, $operation) {
    if ($plugin instanceof PluginWithFormsInterface) {
      if ($plugin->hasFormClass($operation)) {
        return \Drupal::service('plugin_form.factory')->createInstance($plugin, $operation);
      }
    }
  }

  /*************************************************************************
   * Private methods.
   *************************************************************************/

  /**
   * Build form elements for the "mintable" third party settings.
   */
  private function buildMintableElements(array &$element, FormStateInterface $form_state) {
    $wrapper = 'nft_mint';
    $settings = $form_state->get('third_party_settings') ?? [];

    $element[$wrapper] = [
      '#type' => 'details',
      '#title' => $this->t('NFT Mint settings'),
      '#open' => TRUE,
    ];

    $element[$wrapper . '_mintable'] = [
      '#type' => 'checkbox',
      '#title' => t('Is entity mintable?'),
      '#description' => t('Allow users to use "Mint" tab for this entity/bundle.'),
      '#default_value' => $settings['mint']['mintable'] ?? FALSE,
      '#group' => $wrapper,
      '#weight' => -1,
    ];

    $minter_plugins = array_column($this->minterManager->getDefinitions(), 'title', 'id');
    asort($minter_plugins);

    $element[$wrapper][$wrapper . '_minter'] = [
      '#type' => 'radios',
      '#title' => t('Minter'),
      '#description' => t('Select default minter plugin.'),
      '#default_value' => $settings['mint']['minter'] ?? 'default',
      '#options' => $minter_plugins,
      '#states' => [
        'visible' => [
          ':input[name="' . $wrapper . '_mintable"]' => ['checked' => TRUE],
        ],
        'required' => [
          ':input[name="' . $wrapper . '_mintable"]' => ['checked' => TRUE],
        ],
      ],
      '#group' => $wrapper,
    ];
  }

  /**
   * Build form elements for the "ipfs" third party settings.
   */
  private function buildIpfsElements(array &$element, FormStateInterface $form_state) {
    $wrapper = 'nft_ipfs';
    $settings = $form_state->get('third_party_settings') ?? [];

    $element[$wrapper] = [
      '#type' => 'details',
      '#title' => $this->t('IPFS storage'),
      '#open' => TRUE,
    ];

    $element[$wrapper . '_auto'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Automatically upload metadata file to IPFS'),
      '#description' => $this->t("If yes, some of your entities will be published to IPFS automatically.") . '<br />' .
      $this->t('Make sure this is what you want.') . '<br />' .
      $this->t('You do NOT want to enable this if your data are somehow private.'),
      '#default_value' => $settings['ipfs']['auto'] ?? FALSE,
      '#group' => $wrapper,
      '#weight' => -1,
    ];

    $provider_manager = \Drupal::service('plugin.manager.ipfs_provider');
    $provider_default_value = $form_state->getUserInput()[$wrapper . '_provider'] ?? NULL;
    $provider_default_value = $provider_default_value ?: ($settings['ipfs']['provider'] ?? 'localhost');

    $provider_options = array_column($provider_manager->getDefinitions(), 'title', 'id');
    asort($provider_options);

    $element[$wrapper][$wrapper . '_provider'] = [
      '#type' => 'radios',
      '#title' => $this->t('IPFS provider'),
      '#description' => $this->t('Select a provider to connect this site to IPFS for NFT mint operations.') . '<br>' .
      $this->t('Learn more about IPFS providers and configure default settings @link', [
        '@link' => Link::fromTextAndUrl($this->t('on this page'), Url::fromRoute('help.page', [
          'name' => 'ipfs',
        ], [
          'attributes' => ['target' => '_blank'],
        ]))->toString(),
      ]),
      '#options' => $provider_options,
      '#default_value' => $provider_default_value,
      '#states' => [
        'visible' => [
          'input[name="' . $wrapper . '_auto"]' => ['checked' => TRUE],
        ],
        'required' => [
          'input[name="' . $wrapper . '_auto"]' => ['checked' => TRUE],
        ],
      ],
      '#group' => $wrapper,
    ];

  }

}
