<?php

namespace Drupal\nft_mint\Form;

use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Plugin\PluginFormFactoryInterface;
use Drupal\Core\Plugin\PluginWithFormsInterface;
use Drupal\nft\Event\MintEvent;
use Drupal\nft\Event\NftEvents;
use Drupal\nft\Utility\Metadata;
use Drupal\nft\Utility\Transaction;
use Drupal\nft_mint\EntityTypeInfo;
use Drupal\nft_mint\Minter\MinterManagerInterface;
use Drupal\nft_mint\Minter\MinterPluginInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Contracts\EventDispatcher\EventDispatcherInterface;

/**
 * Allow users to Mint an entity as NFT.
 */
class EntityMintForm extends EntityForm {

  /**
   * The minter plugin;
   *
   * @var [type]
   */
  protected $minter;

  /**
   * A logger instance.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The Minter plugin manager.
   *
   * @var \Drupal\nft_mint\Minter\MinterManagerInterface
   */
  protected $minterManager;

  /**
   * The plugin form factory.
   *
   * @var \Drupal\Core\Plugin\PluginFormFactoryInterface
   */
  protected $pluginFormFactory;

  /**
   * An event dispatcher instance to use for configuration events.
   *
   * @var \Symfony\Contracts\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * Our custom service.
   *
   * @var \Drupal\nft_mint\EntityTypeInfo
   */
  protected $entityTypeInfo;

  /**
   * Constructs a new EntityMintForm object.
   *
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger channel
   * @param \Drupal\nft_mint\Minter\MinterManagerInterface $this->plugin_manager_minter
   *   The minter plugin manager.
   * @param \Symfony\Contracts\EventDispatcher\EventDispatcherInterface
   *   The event dispatcher.
   * @var \Drupal\nft_mint\EntityTypeInfo
   *   The custom info service.
   */
  public function __construct(
    LoggerInterface $logger,
    MinterManagerInterface $plugin_manager_minter,
    PluginFormFactoryInterface $plugin_form_factory,
    EventDispatcherInterface $event_dispatcher,
    EntityTypeInfo $entity_type_info
  ) {
    $this->logger = $logger;
    $this->minterManager = $plugin_manager_minter;
    $this->pluginFormFactory = $plugin_form_factory;
    $this->eventDispatcher = $event_dispatcher;
    $this->entityTypeInfo = $entity_type_info;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('logger.channel.nft'),
      $container->get('plugin.manager.minter'),
      $container->get('plugin_form.factory'),
      $container->get('event_dispatcher'),
      $container->get('nft_mint.entity_type_info'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);
    $entity = $this->entity;

    $wrapper_id = $this->getFormId() . '_settings';
    $minter_id = $this->getMinterId($form_state);
    $minter_settings = $form_state->get('minter_settings') ?? [];
    $minter = $this->minterManager->createInstance($minter_id, $minter_settings);

    $minter_options = array_column($this->minterManager->getDefinitions(), 'title', 'id');
    asort($minter_options);

    $form['minter_id'] = [
      '#type' => 'select',
      '#title' => $this->t('Minter'),
      '#options' => $minter_options,
      '#ajax' => [
        'callback' => [get_class($this), 'ajaxRefresh'],
        'wrapper' => $wrapper_id,
        'disable-refocus' => TRUE,
      ],
      '#default_value' => $minter_id,
    ];

    // Build plugin form.
    $form['#tree'] = TRUE;
    $form['#parents'] = [];
    $form['settings'] = [
      '#prefix' => '<div id="' . $wrapper_id . '">',
      '#suffix' => '</div>',
      '#parents' => [],
    ];

    if ($plugin_form = $this->getPluginForm($minter)) {
      $subform_state = SubformState::createForSubform($form['settings'], $form, $form_state);
      $subform_state->set('entity', $entity);
      $form['settings'] = $plugin_form->buildConfigurationForm($form['settings'], $subform_state);
    }

    return $form;
  }

  /**
   * {@inheritDoc}
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    $actions = [];

    $actions['mint'] = [
      '#type' => 'submit',
      '#value' => $this->t('Mint'),
      '#submit' => ['::mint'],
    ];

    $actions['cancel'] = [
      '#type' => 'submit',
      '#value' => $this->t('Cancel'),
      '#submit' => ['::back'],
    ];

    return $actions;
  }

  /**
   * Helper.
   */
  protected function getMinterId(FormStateInterface $form_state) {
    $entity_settings = $this->entityTypeInfo->getThirdPartySettings($this->entity);
    $minter_id = $form_state->getUserInput()['minter_id'] ?? NULL;
    $minter_id = $form_state->getValue('minter_id') ?? NULL;
    $minter_id = $minter_id ?: ($entity_settings['mint']['minter'] ?? NULL);
    $minter_id = $minter_id ?: $this->minterManager->getFallbackPluginId();

    return $minter_id;
  }

  /**
   * Get a minter plugin form.
   */
  public function getPluginForm(MinterPluginInterface $minter) {
    if ($minter instanceof PluginWithFormsInterface) {
      if ($minter->hasFormClass('mint')) {
        return $this->pluginFormFactory->createInstance($minter, 'mint');
      }
    }
  }

  /**
   * Ajax callback.
   */
  public static function ajaxRefresh(array $form, FormStateInterface $form_state) {
    return $form['settings'];
  }

  /**
   * {@inheritDoc}
   */
  protected function copyFormValuesToEntity(EntityInterface $entity, array $form, FormStateInterface $form_state) {
    // Disabled because useless in this case.
    return;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $parents = $form_state->getTriggeringElement()['#parents'];
    $trigger = reset($parents) ?? NULL;

    // No validation on Ajax refresh or if user canceled.
    if (in_array($trigger, ['minter_id', 'cancel'])) {
      return;
    }

    // Update plugin with submitted values.
    $subform_state = SubformState::createForSubform($form['settings'], $form, $form_state);
    $minter_settings = $subform_state->getValues();
    $minter_id = $this->getMinterId($form_state);
    $minter = $this->minterManager->createInstance($minter_id, $minter_settings);

    if ($plugin_form = $this->getPluginForm($minter)) {
      $plugin_form->validateConfigurationForm($form['settings'], $subform_state);
      if (!$subform_state->hasAnyErrors()) {
        $plugin_form->submitConfigurationForm($form['settings'], $subform_state);
      }
    }

    foreach ($subform_state->getErrors() as $error) {
      $form_state->setErrorByName('minter_id', $error);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function mint(array $form, FormStateInterface $form_state) {
    try {
      $subform_state = SubformState::createForSubform($form['settings'], $form, $form_state);
      $minter_settings = $subform_state->getValues();
      $minter_id = $this->getMinterId($form_state);
      $minter = $this->minterManager->createInstance($minter_id, $minter_settings);

      $context = $minter->getConfiguration();
      $context['form'] = $this->getFormId();
      $context['minter'] = $minter->getPluginId();
      $context['entity'] = $this->entity;

      $metadata = $minter->metadata($this->entity, $context);

      // Get file url from plugin's owns code.
      // This will publish metadata as JSON to IPFS, if possible.
      if ($file_url = $minter->toUrl($metadata, $context)) {
        $metadata->set('file_url', $file_url->toString());
      }

      $transaction = $minter->mint($metadata);
    } catch (Error $e) {
      // Create a fake failed transaction if failure.
      $transaction = new Transaction([
        'status' => FALSE,
        'message' => $e->getMessage(),
        'metadata' => $metadata,
        'context' => $context,
      ]);
    }

    $tx_values = $transaction->getValues();
    $message = $this->t('Mint transaction @status<br>@message', [
      '@status' => ($tx_values['status'] ?? NULL) ? 'OK' : 'KO',
      '@message' => $tx_values['error'] ?? $tx_values['message'] ?? '',
    ]);
    if ($tx_values['status'] ?? NULL) {
      $this->messenger()->addMessage($message);
    } else {
      $this->messenger()->addError($message);
    }

    $this->eventDispatcher->dispatch(new MintEvent($transaction), NftEvents::MINT);

    $this->back($form, $form_state);
  }

  /**
   * Cancel callback.
   */
  public function back(array $form, FormStateInterface $form_state) {
    if ($this->entity->hasLinkTemplate('canonical')) {
      $form_state->setRedirectUrl($this->entity->toUrl('canonical'));
    }
  }

}
