<?php

namespace Drupal\nft_mint\Plugin\Minter;

use Drupal\Component\Plugin\PluginBase;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Plugin\PluginWithFormsTrait;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\nft\Utility\Metadata;
use Drupal\nft\Utility\Transaction;
use Drupal\nft_mint\Minter\MinterPluginInterface;
use GuzzleHttp\ClientInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Base class for Minter plugins.
 */
abstract class MinterBase extends PluginBase implements MinterPluginInterface, ContainerFactoryPluginInterface {

  use StringTranslationTrait;
  use PluginWithFormsTrait;

  /**
   * A logger instance.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The HTTP client to fetch the feed data with.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * The serialization class to use.
   *
   * @var \Drupal\Component\Serialization\SerializationInterface
   */
  protected $serializer;

  /**
   * The custom service.
   *
   * @var \Drupal\nft_mint\EntityTypeInfo
   */
  protected $entityTypeInfo;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->setConfiguration($configuration);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = new static($configuration, $plugin_id, $plugin_definition);
    $instance->logger = $container->get('logger.channel.nft');
    $instance->moduleHandler = $container->get('module_handler');
    $instance->httpClient = $container->get('http_client');
    $instance->serializer = $container->get('serializer');
    $instance->entityTypeInfo = $container->get('nft_mint.entity_type_info');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return array_filter((new Metadata())->getValues(), function ($key) {
      return !in_array($key, ['entities']);
    }, ARRAY_FILTER_USE_KEY) + $this->customConfiguration();
  }

  /**
   * Easily identitfy child plugin's custom config for easier form theming.
   *
   * @return array
   *   The custom plugin's configuration.
   */
  public function customConfiguration() {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function getConfiguration() {
    return $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    $this->configuration = $configuration + $this->defaultConfiguration();
  }

  /**
   * Allow plugins to edit form element for their configuration.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   A form state object.
   *
   * @return array
   *   A form element, as render array.
   */
  public function getConfigurationFormElements(FormStateInterface $form_state) {
    $elements = [];

    foreach (array_keys($this->getConfiguration()) as $key) {
      // Default textfield fallback.
      $elements[$key] = [
        '#type' => 'textfield',
        '#title' => ucfirst(str_replace('_', ' ', $key)),
      ];

      // Improved file_url UX.
      if ($key == 'file_url') {
        unset($elements[$key]['#title']);

        $elements[$key] += [
          '#title' => $this->t('File URL'),
          '#placeholder' => 'https://ipfs.io/ipfs/<cid>',
          '#description' => $this->t('You can use a public IPFS url - @link for instance.', [
            '@link' => Link::fromTextAndUrl(
              $this->t('Search for a provider here'),
              Url::fromUri('https://duckduckgo.com/?q=upload+file+to+IPFS', [
                'attributes' => ['target' => '_blank'],
              ])
            )->toString(),
          ]),
        ];
        continue;
      }
    }

    return $elements;
  }

  /**
   * {@inheritdoc}
   */
  public function executeMultiple(array $entities) {
    foreach ($entities as $entity) {
      $this->execute($entity);
    }
  }

  /**
   * {@inheritDoc}
   */
  public function execute(EntityInterface $entity = NULL) {
    try {
      $metadata = $this->metadata($entity);
      $transaction = $this->mint($metadata);
      return $transaction;
    } catch (\Exception$e) {
      $this->logger->error($e->getMessage());
      return FALSE;
    }
  }

  /**
   * {@inheritDoc}
   */
  public function metadata(EntityInterface $entity, array $context = []): Metadata {
    $format = 'json';
    $context['minter'] = $this->getPluginId();
    $values = $this->serializer->normalize($entity, $format, $context);
    return new Metadata($values);
  }

  /**
   * {@inheritDoc}
   */
  public function toUrl(Metadata $metadata, array $contet = []): Url {
    // -------------------------------
    // Using IPFS storage for Metadata
    // -------------------------------
    // Important note about IPFS storage for Metadata:
    // It is considered good practice is to publish your NFTs' metadata as a file
    // on IPFS network so that it becomes `permanent` and `public`, forever.
    // If your NFTs are somehow private, please disable options in settings.
    // Alternative option is to use your own localhost IPFS provider instance.
    if ($this->moduleHandler->moduleExists('ipfs')) {
      $settings = $this->entityTypeInfo->getThirdPartySettings($entity) ?? [];
      if ($settings['ipfs']['auto'] ?? FALSE) {
        $ipfs_provider_manager = \Drupal::service('plugin.manager.ipfs_provider');

        // Get IPFS provider ID.
        // 1- From this plugin's annotation.
        // 2- From third party settings.
        // 3- Fallback to default.
        $ipfs_provider_id = $this->getPluginDefinition()['ipfs_provider'] ??
        $settings['ipfs']['provider'] ??
        $ipfs_provider_manager->getFallbackPluginId('broken');

        $ipfs = $ipfs_provider_manager->createInstance($ipfs_provider_id);

        $cid = $ipfs->add(
          $entity->uuid() . '.json',
          Json::encode($metadata->getValues())
        );
        $file_url = $ipfs->toUrl($cid);
        if ($file_url instanceof Url) {
          return $file_url;
        }

        $this->logger->error($this->t('Error during upload of metadata to IPFS'));
      }
    }

    // Return original Url, if possible.
    if ($current_url = $metadata->getValues()['file_url'] ?? NULL) {
      return Url::fromUri($current_url);
    }

    // Return link to the entity.
    $entity = $context['entity'];
    if ($entity instanceof EntityInterface && $entity->hasLinkTemplate('canonical')) {
      return $entity->toUrl('canonical', ['query' => ['q' => 'metadata']])->setAbsolute(TRUE);
    }

    throw new \Exception('Impossible to get a metadata file url: @metadata', [
      '@metadata' => Json::encode($metadata->getValues()),
    ]);
  }

  /**
   * {@inheritDoc}
   */
  public function mint(Metadata $metadata, array $context = []): Transaction {
    if (!Metadata::isValid($metadata->getValues())) {
      throw new \Exception($this->t('Invalid metadata: @values', [
        '@values' => Json::encode($metadata->getValues()),
      ]));
    }

    return new Transaction([
      'status' => FALSE,
      'message' => $this->t('Minter plugin `@plugin` does not actually mint.<br>Use another plugin or implements your own', [
        '@plugin' => $this->getPluginId(),
      ]),
      'metadata' => $metadata,
      'context' => $context,
    ]);
  }

}
