<?php

namespace Drupal\nft_mint\Plugin\Minter;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Form\FormStateInterface;
use Drupal\nft\Utility\Metadata;
use Drupal\nft\Utility\Transaction;
use Drupal\nft_mint\Plugin\Minter\MinterBase;
use Web3p\EthereumTx\Transaction as Tx;
use Web3\Contract;
use Web3\Web3;

/**
 * Defines a minter plugin ready to talk with an ERC721 smart contract.
 *
 * @Minter(
 *   id = "erc721",
 *   title = @Translation("ERC721"),
 *   description = @Translation("Use an ERC721-compatible contract to mint NFTs."),
 *   forms = {
 *     "mint" = "\Drupal\nft_mint\PluginForm\MinterBaseForm",
 *   }
 * )
 */
class MinterERC721 extends MinterBase {

  /**
   * Web3 provider instance.
   *
   * @var \Web3\Web3
   */
  protected $web3;

  /**
   * The deployed contract.
   *
   * @var \Web3\Contract
   */
  protected $contract;

  /**
   * Separate function to easily identify custom config elements.
   */
  public function customConfiguration() {
    return [
      'network_url' => '',
      'contract_address' => '',
      'contract_abi' => [],
      'account_address' => '',
    ];
  }

  /**
   * {@inheritDoc}
   */
  public function getConfigurationFormElements(FormStateInterface $form_state) {
    $elements = parent::getConfigurationFormElements($form_state);

    $elements['contract_abi']['#type'] = 'textarea';
    $elements['contract_abi']['#rows'] = '3';
    $elements['contract_abi']['#value'] = $form_state->getValue('contract_abi');

    $custom_settings = $this->customConfiguration();
    foreach (array_keys($custom_settings) as $key) {
      $elements[$key]['#required'] = TRUE;
    }

    return $elements;
  }

  /**
   * Connect to the deployed contract.
   *
   * @return $this
   */
  public function connect() {
    if (!$this->contract instanceof Contract) {
      $configuration = $this->getConfiguration();
      $network_url = $configuration['network_url'] ?? '';
      $contract_abi = Json::decode($configuration['contract_abi'] ?? []);

      $this->web3 = new Web3($network_url);
      $this->contract = new Contract($this->web3->provider, $contract_abi);
    }

    return $this;
  }

  /**
   * {@inheritDoc}
   */
  public function mint(Metadata $metadata, array $context = []): Transaction {
    // Make sure Contract is set.
    $this->connect();

    // File URL is not required by default in Metadata. Stop if missing.
    // @see \Drupal\nft_metadata\Normalizer\MetadataNormalizer::normalize();
    if (!$token_uri = $metadata->getValues()['file_url'] ?? NULL) {
      throw new \Exception('Missing file_url to mint ERC721 NFT.');
    }

    $configuration = $this->getConfiguration();
    $network_url = $configuration['network_url'] ?? '';
    $contract_abi = Json::decode($configuration['contract_abi'] ?? []);
    $account_address = $configuration['account_address'] ?? NULL;
    $contract_address = $configuration['contract_address'] ?? NULL;
    $contract_function = $configuration['contract_function'] ?? 'mintEntity';

    // Mint token now.
    $call_data = NULL;
    $this->contract->at($contract_address)->call($contract_function, $token_uri, function ($e, $data) use (&$call_data) {
      if ($e) {
        throw new \Exception($e->getMessage());
      }

      $call_data = $data;
    });

    $transaction = new Tx([
      'nonce' => '0x01',
      'from' => $account_address,
      'to' => $contract_address,
      'gas' => '0x76c0',
      'gasPrice' => '0x9184e72a000',
      'value' => '0x9184e72a', // web3.toWei(1, "ether")
      'data' => $call_data,
    ]);

    $account_private_key = "";
    $signed_tx = $transaction->sign($account_private_key);

    return $signed_tx;
  }

  /**
   * Get ERC721 token informatiom from contract.
   *
   * @return array
   *
   * @throws \Exception
   */
  public function getToken() {
    $this->connect();
    $configuration = $this->getPluginDefinition();
    $contract_address = $configuration['contract_address'] ?? NULL;

    $nft = [];
    $this->contract->at($contract_address)->call('name', [], function ($e, $data) use (&$nft) {
      if ($e) {throw new \Exception($e->getMessage());}
      $nft['name'] = reset($data);
    });
    $this->contract->at($contract_address)->call('symbol', [], function ($e, $data) use (&$nft) {
      if ($e) {throw new \Exception($e->getMessage());}
      $nft['symbol'] = reset($data);
    });

    // Save token information.
    $this->tokens = [$nft];

    return $this->tokens;
  }
}
