<?php

namespace Drupal\nft_mint\PluginForm;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\PluginFormBase;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\nft\Utility\Metadata;

/**
 * A base form class for Minter plugin "mint" operation.
 *
 * Provide a UX to describe an entity's metadata before minting.
 *
 * @see https://docs.opensea.io/docs/metadata-standards#metadata-structure
 */
class MinterBaseForm extends PluginFormBase {

  use StringTranslationTrait;

  /**
   * Return the plugin instance.
   *
   * @return MinterPluginInterface
   */
  public function getPlugin() {
    return $this->plugin;
  }

  /**
   * {@inheritDoc}
   */
  public function getFormId() {
    $form_id = [];
    $form_id[] = 'minter';
    $form_id[] = 'plugin';
    if ($plugin_id = $this->plugin->getPluginId() ?? NULL) {
      $form_id[] = $plugin_id;
    }
    $form_id[] = 'form';

    return implode('_', $form_id);
  }

  /**
   * {@inheritDoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $entity = $form_state->get('entity');
    assert($entity instanceof EntityInterface);
    if (!$entity instanceof EntityInterface) {
      throw new \Exception('Missing entity reference.');
    }

    $context = [];
    $context['form'] = $this->getFormId();
    $context['plugin'] = $this->plugin->getPluginId();
    $metadata = $this->plugin->metadata($entity, $context);
    $metadata_values = $metadata->getValues();

    $form['metadata'] = [
      '#type' => 'details',
      '#title' => $this->t('Metadata'),
      '#open' => TRUE,
      // Place metadata after custom config by default.
      '#weight' => 1,
    ];

    // Build form elements from configuration.
    $elements_settings = $this->plugin->getConfigurationFormElements($form_state);
    foreach ($this->plugin->getConfiguration() as $key => $value) {
      $default_value = $value ?? $metadata_values[$key] ?? NULL;

      $default_settings = [
        '#type' => 'textfield',
        '#title' => $key,
        '#default_value' => $default_value,
        '#required' => Metadata::isRequired($key),
        '#group' => 'metadata',
      ];

      // Allow children plugins to configure form elements.
      // @see \Drupal\nft_mint\Plugin\Minter\MinterBase::getConfigurationFormElements();
      $custom_settings = $elements_settings[$key] ?? [];

      $form[$key] = array_merge($default_settings, $custom_settings);
    }

    $plugin_id = $this->plugin->getPluginId();
    $custom_config = $this->plugin->customConfiguration();

    $form[$plugin_id] = [
      '#type' => 'details',
      '#title' => $this->plugin->getPluginDefinition()['title'] ?? $this->t('Custom settings'),
      '#open' => TRUE,
      '#access' => !empty($custom_config),
    ];
    foreach (array_keys($custom_config) as $key) {
      $form[$key]['#group'] = $plugin_id;
    }

    return $form;
  }

  /**
   * {@inheritDoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    // Nothing specific to validate.
  }

  /**
   * {@inheritDoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $plugin_keys = array_keys($this->plugin->getConfiguration());
    $settings = array_filter($form_state->getValues(), function ($key) use ($plugin_keys) {
      return in_array($key, $plugin_keys);
    }, ARRAY_FILTER_USE_KEY);
    $this->plugin->setConfiguration($settings);
  }

}
